/*
BeTiled!
Mi primer programa en C para Sinclair Spectrum

Luis I. Garcia / Benway

(c) 2.007 ComputerEmuZone Games Studio
*/

#include "main.h"
  
void SetRAMBank()
{
#asm	
    .SetRAMBank
	ld  A, ($5B5C)
	and F8h
	or  B
	ld  BC, $7FFD
	ld  ($5B5C), A
	out (C), A
#endasm
}

/* MusicISR */
#asm
defw 0                        // 2 bytes libres
#endasm
void ISR(void)
{
    if (is128k) {
	   #asm
	        ld b,1
	        call SetRAMBank
			call WYZPLAYERISR
	        ld b,0
	        call SetRAMBank
	   #endasm
       }
    if (++Ticks == 50) {
       Ticks = 0;
       Segs++;
       }
    return_nc(0);
}
/* Fin de ISR */

void wyz_play_sound (uchar fx_number)
{
    if (!is128k)
       return;
    asm_int = 2 + (fx_number /*<< 8*/ * 256); 
	#asm
	    di
        ld b,1
        call SetRAMBank
        ld bc, (_asm_int)
		call INICIAEFECTO
        ld b,0
        call SetRAMBank
        ei
	#endasm
}

void wyz_play_music (uchar song_number)
{
     if (!is128k)
        return;
     asm_int = song_number;
     
     #asm
         di
         ld b, 1
         call SetRAMBank
         ld a, (_asm_int)
         call CARGA_CANCION
         ld b, 0
         call SetRAMBank
         ei
     #endasm
}

void wyz_stop_sound ()
{
    if (!is128k)
       return;
    #asm
    di
    ld b,1
    call SetRAMBank
    call SILENCIA_PLAYER
    ld b,0
    call SetRAMBank
    ei
    #endasm
}     

void fade_out ()
{
// El segundo programa de ms de 5 lineas en Z80 ASM de na_th_an
/* Optimizado por Metalbrain */ 
/* Esto "oscurece" un paso la pantalla. Llamado 8 veces la dejar en negro */ 

#asm 
  ld  e, 3            ; 3 tercios 
  ld  hl, 22528       ; aqu empiezan los atributos 
  halt                ; esperamos retrazo. 
.fade_out_bucle 
  ld  a, (hl)         ; nos traemos el atributo actual 

  ld  d, a            ; tomar atributo 
  and  7              ; aislar la tinta 
  jr  z, ink_done     ; si vale 0, no se decrementa 
  dec  a              ; decrementamos tinta 
.ink_done 
  ld  b, a            ; en b tenemos ahora la tinta ya procesada. 

  ld  a, d            ; tomar atributo 
  and  56             ; aislar el papel, sin modificar su posicin en el byte 
  jr  z, paper_done   ; si vale 0, no se decrementa 
  sub 8               ; decrementamos papel restando 8 
.paper_done 
  ld  c, a            ; en c tenemos ahora el papel ya procesado. 
  ld  a, d 
  and  192            ; nos quedamos con bits 6 y 7 (BRIGHT y FLASH) 
  or  c               ; aadimos paper 
  or  b               ; e ink, con lo que recompuesto el atributo 
  ld  (hl),a          ; lo escribimos, 
  inc  l              ; e incrementamos el puntero. 
  jr  nz, fade_out_bucle ; continuamos hasta acabar el tercio (cuando L valga 0) 
  inc  h              ; siguiente tercio 
  dec  e 
  jr  nz, fade_out_bucle ; repetir las 3 veces 
#endasm 
}

void Fade ()
{
     sp_MoveSprAbs (Selector, Tablero, 0, 22,32,0,0);
     
     #asm
         ld b, 7
         .bucle_TotalFade
         push bc
         call _fade_out
         pop bc
         djnz bucle_TotalFade
     #endasm

     sp_UpdateNow();
}     

uint rand ()
{
#asm
.rand

   ld   hl,(_rand_int)
   ld   a,h
   add  a,a                     ;Set highest bit of seed to carry
   rl   l                       ;rotate L left (C<=L<=C)
   rl   h                       ;rotate H left (C<=L<=C)
   add  a,a                     ;Set second highest bit of seed to carry
   rl   l
   rl   h
   add  a,a
   rl   l
   rl   h
   ld   bc,$7415
   add  hl,bc                   ;Add $7415 to HL
   ld   (_rand_int),hl
   ret
/*
._Rand8	ld	a,0		; Seed is usually 0
		ld	b,a
		add	a,a
		add	a,a
		add	a,b
		inc	a		; another possibility is ADD A,7
		ld (_Rand8+1),a
		ld (_rand_int), a*/
#endasm

	return rand_int;
}

unsigned int abs (int x)
{
	if (x >= 0)
		return x;
	else
		return -x;
}

void uncrunch (uint address)
{
    asm_int = address;

	#asm
	    ld hl, 22528       ; Primero ponemos los atributos en negro
	    ld (hl), 0         ; para que no se vea feo al descomprimir
	    ld d, h
	    ld e, l
	    inc de
	    ld bc, 767
	    ldir
	    
	    ld de, $4000       ; Vamos a descomprimir en la pantalla

	    ld hl, (_asm_int)  ; Lo que nos hayan pedido
	    
		call PUCRUNCH
	#endasm
}

void InitCharset()
{
     for (n = 48; n < 58; n++)  // Los nmeros
         sp_TileArray (n, charset16 + ((n - 48) * 8));

     for (n = 65; n < 91; n++)  // Las maysculas
         sp_TileArray (n, charset65 + ((n - 65) * 8));
         
     for (n = 120; n < 160; n++)  // La Explosin de los tiles y los tiles Especiales
         sp_TileArray (n, boom0 + ((n - 120) * 8));
}

void InitTileset()
{
     uchar *PSet;
     
     PSet = tileset1_0 + 224 * TileSet;  // 224 bytes ocupa cada tileset
     
     for (n = 92; n < 120; n++) // Los tiles en s
         sp_TileArray (n, PSet + ((n - 92) * 8));
}

void UnpackScreenInPage (uchar Page, uint Addr)
{
/* Esta funcin es IMPRESCINDIBLE porque as puedo descomprimir datos (pantallas en este caso) que
estn en otra pgina de RAM sin problemas, INDEPENDIENTEMENTE de desde donde lo quiera hacer (Esta funcin
est al principio del programa, y, por lo tanto, no se pierde su cdigo al paginar.
Es decir, sin esto, si cojo y meto este cdigo funcionar si est por debajo de $C000, pero si lo hago despus,
cuando retorne desde el call SetRAMBank, no se va a encontrar lo que haba antes, sino lo que hubiera en la
pgina que hemos cargado */

     asm_int = Page;

     #asm
         di
         ld a,(_asm_int)
         ld b,a
         call SetRAMBank
     #endasm
     uncrunch (Addr);
     #asm
         ld b,0
         call SetRAMBank
         ei
     #endasm
}
     

void Init ()
{
    sp_Border (BLACK);
    sp_Initialize (INK_BLACK | PAPER_BLACK, ' ');
    sp_UpdateNow();

    #asm
        di
    #endasm

    sp_InitIM2 (0xf2f2);  // f1f1
    sp_CreateGenericISR (0xf2f2);  // f1f1
    sp_RegisterHookLast (255, ISR);

    #asm
        LD   HL,$F1FF // Chequea si son 128k, y lo pone en is128k
        LD   (HL),1
        LD   A,$01
        LD   BC,7FFDh
        OUT   (C),A
        LD   (HL),0
        LD  A,($5B5C)
        AND $F8
        LD   BC,$7FFD
        OUT   (C),A
        LD   A,(HL)
        LD (_is128k),A
        ei
    #endasm

	sp_AddMemory(0, 24, 14, AD_FREE);

    Selector = sp_CreateSpr (sp_MASK_SPRITE, 3, sel1, 1, TRANSPARENT);
    sp_AddColSpr (Selector, sel2, TRANSPARENT);
    sp_AddColSpr (Selector, sel3, TRANSPARENT);

    keys.up = sp_LookupKey ('q');
    keys.down = sp_LookupKey ('a');
    keys.left = sp_LookupKey ('o');
    keys.right = sp_LookupKey ('p');
    keys.fire = sp_LookupKey (' ');
    
    *sp_KeyStartRepeat = 255;      // Con esto "desactivamos" la repeticin de teclas ... excepto para los "Joysticks"

    HiScore = 0;   
    Mode = 0;
    Locker = 0;
    Control = 0;
    TileSet = 0;

    InitCharset();
    InitTileset();
    
    if (is128k) {
       UnpackScreenInPage (3, CEZSILVER);
       sp_Pause(150);
       Fade();
       UnpackScreenInPage (3, MOJONTWINS);
       sp_Pause(150);
       Fade();
       }
       
}

void PrintAtNum (uchar y, uchar x, uchar PColor, uint Num)
{
    if (!Num)
       sp_PrintAtInv (y, x, PColor, '0');

    while (Num) {
          sp_PrintAtInv (y, x--, PColor, '0' + (Num % 10));
          Num /= 10;
          }
}

void PrintAtStr (uchar y, uchar x, uchar PColor, uchar *Str)
{
     uchar Bucle = 0;
     while (Str[Bucle])
           sp_PrintAtInv (y, x + Bucle, PColor, Str[Bucle++]);
}

void ShowTime ()
{
     PrintAtStr (10, 25, 2, "00");
     if (Time - Segs > TIMELIMIT)
        return;
     if (Time - Segs > 9)
        PrintAtNum (10, 26, 7, Time - Segs);
      else {
        PrintAtNum (10, 26, 2, Time - Segs);
        }
}     

void ShowScore ()
{
     CheckEnd = 0;
     for (n = 1; n < 8; n++) {
         Color = PAPER_BLACK | INK_WHITE;
         if (TilesD[n - 1] >= Objetivo) {
            Color = PAPER_WHITE | INK_BLACK;
            CheckEnd++;
            }
         sp_PrintAtInv (21, 3 * (n - 1) + 1, Color, '0'); 
         PrintAtNum (21, 3 * (n - 1) + 2, Color, TilesD[n - 1]);
         }
     if (Score > HiScore)
        HiScore = Score;
     PrintAtNum (4, 28, PAPER_BLACK | INK_WHITE, Score);
     PrintAtNum (7, 28, PAPER_BLACK | INK_WHITE, HiScore);
}

void PrintBanner (uchar y, uchar x, uchar PColor, char *Str)
{
     uchar Blank[32];

     sp_MoveSprAbs (Selector, Tablero, 0, 22,32,0,0);     

     Blank[31] = 0;
     while (Str[Blank[31]])
           Blank[Blank[31]++] = ' ';
     Blank [Blank[31]] = ' ';
     Blank[Blank[31] + 1] = 0;
     
     PrintAtStr (y - 1, x + 1, PColor, Blank);
     PrintAtStr (y, x + 1, PColor, Blank);
     PrintAtStr (y, x + 1, PColor, Str);
     PrintAtStr (y + 1, x + 1, PColor, Blank);
     for (Blank[31] = 0; Blank[31] < 3; Blank[31]++)
         sp_PrintAtInv (y - 1 + Blank[31], x, PColor, ' ');
}

void PrintIndic (uchar PrTime, char *Str1, char *Str2)
{
     PrintBanner (13, 22, Color, "    ");
     PrintBanner (16, 22, Color, "    ");
     PrintAtStr (13, 22, Color, Str1);
     PrintAtStr (16, 22, Color, Str2);
     TimeIndic = Segs + PrTime;
}

void ClearIndic ()
{
    Color = INK_WHITE;
    PrintIndic (0, "  0","  0");

    PrintAtStr (12, 22, PAPER_BLACK | INK_MAGENTA, "LEVEL:"); 
    PrintAtNum (13, 25, PAPER_BLACK | INK_WHITE, Objetivo - 2);
    PrintAtStr (15, 22, PAPER_BLACK | INK_MAGENTA, "OBTAIN"); 
    PrintAtNum (16, 25, PAPER_BLACK | INK_WHITE, Objetivo);
    TimeIndic = 0;
}

void CreaTile (uchar Donde)
{
     if (Mode == 1 || (((uchar)rand()) % 100) > 2 || (Mode == 0 && Objetivo < INTERLEVEL + 3)) {
        TileGrid [Donde] = (((uchar)rand() % 7)) + 1;
        return;
        }
     
     TileGrid [Donde] = (((uchar)rand()) % SPECIALTILE) + 12; // 12 es el primer tile "especial"

     if (TileGrid [Donde] == 18) {
        if (Cambiante < 0)
           Cambiante = Donde;
         else
           TileGrid [Donde] = (((uchar)rand()) % 7) + 1;
        }
}

void DibujaTile (uchar x, uchar y, char Tile, uchar Paper, uchar Flash, char Offset)
{
     if (Tile < 0) 
        Tile = 11; // El 11 est "en blanco"
             
     if (Tile < 8) {
        Color = Tile | Paper;

        if (Paper)
           Color = Paper;

        }
      else
        Color = Paper;

     if (Tile == 18) { // Tile cambiante
        Tile = (((uchar)rand()) % 7) + 1;
        Color = PAPER_BLUE * Tile;
        }

     if (Tile > 11)
        Color = INK_WHITE;

     if (Flash)
        Color = PAPER_BLUE * Color | FLASH | BRIGHT;

     if (Offset + y >= 0) {
        sp_PrintAtInv (ORIGENGRIDY + y * 2 + Offset, ORIGENGRIDX + x * 2, Color | BRIGHT, 92 + (Tile - 1) * 4);
        sp_PrintAtInv (ORIGENGRIDY + y * 2 + Offset, ORIGENGRIDX + x * 2 + 1, Color, (92 + (Tile - 1) * 4) + 1);
        }
     sp_PrintAtInv (ORIGENGRIDY + y * 2 + 1 + Offset, ORIGENGRIDX + x * 2, Color, (92 + (Tile - 1) * 4) + 2);
     sp_PrintAtInv (ORIGENGRIDY + y * 2 + 1 + Offset, ORIGENGRIDX + x * 2 + 1, Color, (92 + (Tile - 1) * 4) + 3);
}

void DrawPreLine (uchar x1, uchar y1, uchar x2, uchar y2, uchar x3, uchar y3, uchar Draw, uchar *Ret)
{
     if (Draw) {
        DibujaTile (x1, y1, TileGrid [y1 * DIMGRID + x1], PAPER_BLACK, 1, 0);
        DibujaTile (x2, y2, TileGrid [y2 * DIMGRID + x2], PAPER_BLACK, 1, 0);
        DibujaTile (x3, y3, TileGrid [y3 * DIMGRID + x3], PAPER_BLACK, 1, 0);
        }
     
     *Ret = 1;
}

char CheckPreLines (uchar Draw)
{
      char *Tile;
      uchar Ret;

      Ret = 0;
      
      Tile = TileGrid;

      for (n = 0; n < DIMGRID; n++)
          for (m = 0; m < DIMGRID; m++, Tile++) {

              if (*Tile > 7) // Si hay una ficha "especial" ya tenemos un movimiento posible
                 DrawPreLine (m, n, m, n, m , n, Draw, &Ret);

              if (m < DIMGRID - 1 && *Tile == *(Tile + 1)) {                       // O O
                 if (m < DIMGRID - 3 && *Tile == *(Tile + 3))                              // O O X O
                    DrawPreLine (m, n, m + 1, n, m + 3, n, Draw, &Ret);
                 if (m > 1 && *Tile == *(Tile - 2))                                        // O X O O
                    DrawPreLine (m, n, m + 1, n, m - 2, n, Draw, &Ret);
                 if (n > 0 && m > 0 && *Tile == *(Tile - DIMGRID - 1))                     // ^ O O
                    DrawPreLine (m, n, m + 1, n, m - 1, n - 1, Draw, &Ret);
                 if (n > 0 && m < DIMGRID - 2 && *Tile == *(Tile + 2 - DIMGRID))           // O O ^
                    DrawPreLine (m, n, m + 1, n, m + 2, n - 1, Draw, &Ret);
                 if (n < DIMGRID - 1 && m > 0 && *Tile == *(Tile + DIMGRID - 1))           // _ O O
                    DrawPreLine (m, n, m + 1, n, m - 1, n + 1, Draw, &Ret);
                 if (n < DIMGRID - 1 && m < DIMGRID - 2 && *Tile == *(Tile + DIMGRID + 2)) // O O _
                    DrawPreLine (m, n, m + 1, n, m + 2, n + 1, Draw, &Ret);
                 }

              if (m < DIMGRID - 2 && *Tile == *(Tile + 2)) {                     // O X O
                 if (n > 0 && *Tile == *(Tile - DIMGRID + 1))                              // O ^ O
                    DrawPreLine (m, n, m + 1, n - 1, m + 2, n, Draw, &Ret);
                 if (n < DIMGRID - 1 && *Tile == *(Tile + DIMGRID + 1))                    // O _ O
                    DrawPreLine (m, n, m + 1, n + 1, m + 2, n, Draw, &Ret);
                 }

              if (n < DIMGRID - 1 && *Tile == *(Tile + DIMGRID)) {              // O O
                 if (n < DIMGRID - 3 && *Tile == *(Tile + (3 * DIMGRID)))                    // O O X O
                    DrawPreLine (m, n, m, n + 1, m, n + 3, Draw, &Ret);
                 if (n > 1 && *Tile == *(Tile - (2 * DIMGRID)))                              // O X O O
                    DrawPreLine (m, n, m, n + 1, m, n - 2, Draw, &Ret);
                 if (m > 0 && n > 0 && *Tile == *(Tile - DIMGRID - 1))                     // < O O
                    DrawPreLine (m, n, m, n + 1, m - 1, n - 1, Draw, &Ret);
                 if (m > 0 && n < DIMGRID - 2 && *Tile == *(Tile + (2 * DIMGRID) - 1))       // O O <
                    DrawPreLine (m, n, m, n + 1, m - 1, n + 2, Draw, &Ret);
                 if (m < DIMGRID - 1 && n > 0 && *Tile == *(Tile - DIMGRID + 1))               // > O O
                    DrawPreLine (m, n, m, n + 1, m + 1, n - 1, Draw, &Ret);
                 if (m < DIMGRID - 1 && n < DIMGRID - 2 && *Tile == *(Tile + (2 * DIMGRID) + 1 ))// O O >
                    DrawPreLine (m, n, m, n + 1, m + 1, n + 2, Draw, &Ret);
                 }

              if (n < DIMGRID - 2 && *Tile == *(Tile + 2 * DIMGRID)) {          // O X O
                 if (m < DIMGRID - 1 && *Tile == *(Tile + DIMGRID + 1))                    // O > O
                    DrawPreLine (m, n, m + 1, n + 1, m, n + 2, Draw, &Ret);
                 if (m > 1 && *Tile == *(Tile + DIMGRID - 1))                              // O < O
                    DrawPreLine (m, n, m - 1, n + 1, m, n + 2, Draw, &Ret);
                 }
              }
     return Ret;
}

void ResetGrid ()
{
    char *Tile;
    
    Cambiante = -1;

    for (n = 0; n < DIMGRID * DIMGRID; n++)
        CreaTile(n);

    do {
        p = 0;
        for (n = 1; n < DIMGRID - 1; n++)
            for (m = 0; m < DIMGRID; m++) {
                Tile = &TileGrid [n * DIMGRID + m];
                if (*Tile == *(Tile - DIMGRID) &&
                    *Tile == *(Tile + DIMGRID)) {
                    if (++(*Tile) > 7)
                       *Tile = 1;
                    p = 1;
                    }

                Tile = &TileGrid [m * DIMGRID + n];
                if (*Tile == *(Tile - 1) &&
                    *Tile == *(Tile + 1)) {
                    if (++(*Tile) > 7)
                       *Tile = 1;
                    p = 1;
                    }
            }
        } while (p || !CheckPreLines (0));
}

void DrawGrid ()
{
    for (n = 0; n < DIMGRID * DIMGRID; n++)
        DibujaTile (n % DIMGRID, n / DIMGRID, TileGrid[n], PAPER_BLACK, 0, 0);
    ShowScore();
}

void Gravedad ()
{
     for (m = 0; m < DIMGRID; m++)
         for (n = DIMGRID; n > 0; n--) {
             if (TileGrid [(n-1) * DIMGRID + m] <= 0) {
                for (o = (n-1); o > 0; o--)
                    TileGrid [o * DIMGRID + m] = TileGrid [(o - 1) * DIMGRID + m];
                CreaTile(m);
                for (p = 2; p > 0; p--) {
                    for (o = n; o > 0; o--)
                        DibujaTile (m, (o-1), TileGrid [(o-1) * DIMGRID + m], PAPER_BLACK, 0, - (p - 1));
                    sp_UpdateNow();
                    }
                n++;
                }
         }
}

void Explosion (uchar Wait)
{
     char *Tile;
     for (o = 0; o < 4; o++) {
         Tile = TileGrid;
         for (n = 0; n < DIMGRID * DIMGRID; n++, Tile++)
             if (*Tile < 0)
                DibujaTile (n % DIMGRID, n / DIMGRID, 8 + o, INK_BLACK + abs(*Tile), 0, 0);
         sp_UpdateNow();
         if (Wait)
            sp_Wait (2);  // Retardo para que se vea la explosin
         }
}     

char FindLines ()
{
     char *Tile;
     uchar PreScore, Lines;
     
     for (Tile = TileGrid, Lines = 0, n = 0; n < DIMGRID; n++)
         for (m = 0; m < DIMGRID; m++, Tile++) {
                                      
              if (*Tile < 8 && m < DIMGRID - 2 &&
                  abs(*Tile) == abs (*(Tile + 1)) &&
                  abs(*Tile) == abs(*(Tile + 2))) {
                  if (*Tile > 0 || *(Tile + 1) > 0 || *(Tile + 2) > 0) {
                     Lines++;
                     wyz_play_sound (FX_LINE);
                     }

                  o = 0;
                  PreScore = 0;
                  while (m + o < DIMGRID && abs(*Tile) == abs(*(Tile + o))) {
                        if (*(Tile + o) > 0) {
                            *(Tile + o) *= -1;
                            TilesD [abs(*Tile) - 1]++;
                            PreScore++;
                            }
                        o++;
                        }
                  if (PreScore) {
                     Score += PreScore - 2;

                     if (PreScore > 3) {
                        Color = PAPER_CYAN | INK_BLACK;
                        PrintIndic (3, "  X", " LINE");
                        wyz_play_sound (FX_45);
                        sp_PrintAtInv (13, 25, PAPER_CYAN | INK_BLACK, '0' + PreScore);
                        }

                     if (abs(*Tile) == Doble)
                        Score += PreScore - 2;
                     if (++Time > TIMELIMIT + Segs)
                        Time = TIMELIMIT + Segs;
                     }
                  }

              if (*Tile < 8 && n < DIMGRID - 2 &&
                  abs(*Tile)  == abs(*(Tile + DIMGRID)) &&
                  abs(*Tile)  == abs(*(Tile + DIMGRID * 2))) {
                  if (*Tile > 0 || *(Tile + DIMGRID) > 0 || *(Tile + 2 * DIMGRID) > 0) {
                     Lines++;
                     wyz_play_sound (FX_LINE);
                     }
                  o = 0;
                  PreScore = 0;
                  while (n + o < DIMGRID && abs(*Tile) == abs(*(Tile + o * DIMGRID))) {
                        if (*(Tile + o * DIMGRID) > 0) {
                            *(Tile + o * DIMGRID) *= -1;
                            TilesD [abs(*Tile)  - 1]++;
                            PreScore++;
                            }
                        o++;
                        }
                  if (PreScore) {
                     Score += PreScore - 2;

                     if (PreScore > 3) {
                        Color = PAPER_CYAN | INK_BLACK;
                        PrintIndic (3, "  X", " LINE");
                        wyz_play_sound (FX_45);
                        sp_PrintAtInv (13, 25, PAPER_CYAN | INK_BLACK, '0' + PreScore);
                        }

                     if (abs(*Tile)  == Doble)
                        Score += PreScore - 2;
                     if (++Time > TIMELIMIT + Segs)
                        Time = TIMELIMIT + Segs;
                     }
                  }
              }

     if (Lines)
        Explosion(1);
     return Lines;
}

char SwapTile (uchar x, uchar y, char Sent)
{
     uchar Combo, Lines, SSegs;
     char *Tile;
     
     Tile = &TileGrid [DIMGRID * y + x];

     if (Sent) {
        wyz_play_sound (FX_FLIP);
        sp_Swap (Tile, Tile + Sent, 1);
        DibujaTile (x, y, *Tile, PAPER_BLACK, 0, 0);
        DibujaTile (x + Sent % DIMGRID, y + Sent / DIMGRID, *(Tile + Sent), PAPER_BLACK, 0, 0);
        }

     sp_UpdateNow();
     
     Combo = FindLines();

     if (!Combo && Sent) {  // Si no has hecho lneas, y es una jugada de moverse, no de cargarse un tile especial...
        wyz_play_sound (FX_FLOP);
        sp_Swap (Tile, Tile + Sent, 1);
        DibujaTile (x, y, *Tile, PAPER_BLACK, 0, 0);
        DibujaTile (x + Sent % DIMGRID, y + Sent / DIMGRID, *(Tile + Sent), PAPER_BLACK, 0, 0);
        sp_Wait (10); // <------------------------ RETARDO para ver el cambio de tiles
        ShowTime();
        return 0;
        }

     do {
        ShowScore();
        ShowTime();
        Gravedad();
        Lines = FindLines();
        Combo += Lines;
        if (Combo > 1) {
           Color = PAPER_YELLOW | INK_BLACK;
           PrintIndic (3, "COMBO", " X");
           PrintAtNum (16, 25, PAPER_YELLOW | INK_BLACK, Combo);
           Score += Combo * 2;
           ShowScore();
           }
        } while (Lines);
        
        if (Combo > 1)
           wyz_play_sound (FX_COMBO);


     if (CheckEnd < 7 && !CheckPreLines(0)) {
        SSegs = Segs;
        PrintBanner (5, ORIGENGRIDX, PAPER_RED | INK_YELLOW, "NO MORE  LINES");
        sp_UpdateNow();
        sp_Wait (100);
        do {
            ResetGrid();
            } while (!CheckPreLines(0));
        DrawGrid();
        Segs = SSegs;
        Ticks = 0;
        ShowTime();
        sp_UpdateNow();
        }
     
     return 1;
}

void SpecialTile (uchar y, uchar x)
{
     if (TileGrid [y * DIMGRID + x] == 15)
         wyz_play_sound (FX_SPIR);
       else
         wyz_play_sound (FX_ITEM);

     switch (TileGrid [y * DIMGRID + x]) {
            case 12:     // Pesa
                 while (y < DIMGRID - 1) {
                       if (TilesD[TileGrid [(y+1) * DIMGRID + x] - 1])
                          TilesD[TileGrid [(y+1) * DIMGRID + x] - 1]--;
                       TileGrid [++y * DIMGRID + x] *= -1;
                       Score++;
                       ShowScore();
                       Gravedad();
                       }
                 Score++;
                 break;
            case 13:     // Reloj
                 Color = PAPER_GREEN | INK_BLACK;
                 PrintIndic (3, " TIME", "  +5 ");
                 Time += 5;
                 if (Time > TIMELIMIT + Segs)
                    Time = TIMELIMIT + Segs;
                 break;
            case 14:     // Cruz
                 Color = PAPER_RED | INK_WHITE | BRIGHT | FLASH;
                 PrintIndic (5, "INVER-", " TED!"); 
                 TimeR = Segs + 5;
                 Score += 5;
                 break;
            case 15:    // Espiral
                 do {
                    ResetGrid();
                    } while (!CheckPreLines(0));
                 DrawGrid();
                 Score += 5;
                 Time -= 2;
                 ShowTime();
                 sp_UpdateNow();
                 return;
            case 16:    // + 100
                 Score += 10;
                 break;
            case 17:    // - 100
                 if (Score > 10)
                    Score -= 10;
                  else
                    Score = 0;
                 break;
            case 18: // Cambiante
                 p = (((uchar)rand()) % 7) + 1;
                 DibujaTile (x, y, p, PAPER_BLACK, 0, 0);
                 sp_UpdateNow();
                 sp_Wait (5);
                 for (n = 0; n < DIMGRID * DIMGRID; n++)
                     if (TileGrid[n] == p) {
                        TileGrid[n] *= -1;
                        Score++;
                        }
                 Cambiante = -1;
                 break;
            }
    ShowTime();
    ShowScore();
    TileGrid [y * DIMGRID + x] = -7;
    Explosion(1);
    Gravedad();
    SwapTile (0,0,0);
}

void StartLevel ()
{
    Doble = (((uchar)rand()) % 7) + 1;
    TimeR = 0;

    sp_UpdateNow();

    TX = DIMGRID / 2;
    TY = DIMGRID / 2;
    ObjX = TX;
    ObjY = TY;
    XAct = 8 * ORIGENGRIDX + 16 * DIMGRID / 2;
    YAct = 8 * ORIGENGRIDY + 16 * DIMGRID / 2;

    #asm
//    for (Color = 0; Color < 7; Color++)
//        TilesD[Color] = 0;
// Esto hace lo mismo, pero ocupa 32 bytes menos
        ld hl, _TilesD
        ld b,7
        .bucle_ResetTilesD
        ld (hl), 0
        inc hl
        djnz bucle_ResetTilesD
    #endasm

    
    #asm //    while (++Objetivo < 3);  es lo mismo, pero en asm ocupa 5 bytes menos
        ld a, (_Objetivo)
        .bucle_SubObjetivo
        inc a
        cp 3
        jr c, bucle_SubObjetivo
        ld (_Objetivo), a
    #endasm
    
    Swaping = 0;
    FireReleased = 1;
    sp_Validate(sp_ClipStruct, sp_ClipStruct); // Para no borrar el marco
    ResetGrid();
    DrawGrid();
    DibujaTile (9, 7, Doble, PAPER_BLACK, 0, -1);    
    PrintAtStr (3, 24, PAPER_BLACK | INK_MAGENTA, "SCORE");
    PrintAtStr (4, 23, PAPER_BLACK | INK_WHITE, "0000000");
    PrintAtStr (6, 23, PAPER_BLACK | INK_MAGENTA, "HISCORE");
    PrintAtStr (7, 23, PAPER_BLACK | INK_WHITE, "0000000");
    PrintAtStr (9, 24, PAPER_BLACK | INK_MAGENTA, "TIME");   
    PrintAtNum (10, 26, PAPER_BLACK | INK_WHITE, Time);
    ShowScore();

    ClearIndic();
    
    PrintBanner (ORIGENGRIDY + 5, ORIGENGRIDX + 3, PAPER_BLUE | INK_YELLOW, "LEVEL 00");
    PrintAtNum (ORIGENGRIDY + 5, ORIGENGRIDX + 11, PAPER_BLUE | INK_YELLOW, Objetivo - 2);
    sp_UpdateNow();
    
    wyz_play_sound (FX_PAUSA);

    sp_Wait (100);

    DrawGrid();
    sp_UpdateNow();

    Ticks = 0;
    Segs = 0;
}
             
void MyScr ()
{
// Desplaza un pxel a la izquierda la ltima fila de caracteres
#asm
    ld hl, 0x57ff ; Empezamos en la ltima lnea
    ld c, 8       ; 8 pxeles de alto
    halt          ; Sincronizamos con el retrazo
    .buc_2
    ld b,32       ; 32 bytes de ancho
    .buc_1
    rl (hl)       ; Rotamos el byte que corresponda
    dec hl        ; Y pasamos al siguiente
    djnz buc_1
    dec h         ; Para no pasarnos de caracter, tenemos que restar (255 - 32)
    ld a,l        ; Y ahora sumamos el 32 que no restamos antes
    add a,32      ; Pues eso
    ld l,a        ; Y lo devolvemos a L
    dec c
    jr nz, buc_2
#endasm
}

void MenuInfo()
{
     PrintAtStr (10, 6, PAPER_BLACK | INK_RED | BRIGHT, "1 GAME MODE:");

     switch (Mode) {
            case 0:
                PrintAtStr (10, 19, PAPER_BLACK | INK_YELLOW | BRIGHT, "ARCADE ");
                break;
            case 1:
                PrintAtStr (10, 19, PAPER_BLACK | INK_YELLOW | BRIGHT, "NORMAL ");
                break;
            case 2:
                PrintAtStr (10, 19, PAPER_BLACK | INK_YELLOW | BRIGHT, "SPECIAL");
                break;
            }

     if (Mode) {
       PrintAtStr (12, 6, PAPER_BLACK | INK_GREEN | BRIGHT, "2 TILESET:");
       for (n = 0; n < 10; n++) // Este, el tileset, ahorra al hacerlo as porque son 5 cadenas de 10 bytes cada una y el switch es muy largo
         sp_PrintAtInv (12, 17 + n, PAPER_BLACK | INK_YELLOW | BRIGHT, CSets [TileSet * 10 + n]);
       }
      else
       PrintAtStr (12, 6, 0, "                     ");
}

void CheckFireIsPressed ()
{
     if ((((JoyFunc) (&keys))) != 127)
        return;
     
    #asm
        pop af                ; Hemos venido con un call: La ltima direccin en la pila, es la direccin de retorno, pero antes de ella,
                              ; como no creamos variables locales ni nada, z88dk no ha metido nada, luego est la direccin de retorno de
                              ; la funcin anterior, que es Menu().
                              ; Si bajamos la "nuestra", de CheckFireIsPressed(), al volver, volvemos a donde tenga que volver Menu()
                              ; Esto solo va a pasar cuando se pulse el fuego, es decir, hay que salir del men
        call _wyz_stop_sound  
        call _Fade
        ret
    #endasm
     
}

void Menu ()
{
     m = 8; o = 0; asm_int = 0;
     Score = 0;

     Fade();

     uncrunch (MENU);

     MenuInfo();
  
     wyz_play_music (MUSIC_MENU);
         
     while (1) {   // Siempre? No! Ver comentario en funcin CheckFireIsPressed
           PrintAtStr (21, 6, (((uchar)rand()) % 7) + 1, "PRESS FIRE TO START");
           
          p = sp_GetKey();
 
          if (p == 49) {  // Ha pulsado el 1
              wyz_play_sound (FX_TECLA);
              #asm
//              if (++Mode > 2)
//                 Mode = 0;
//              if (Mode > 1 && !Locker)
//                 Mode = 0;
//              MenuInfo();
// En asm ocupa 18 bytes menos
                  ld hl, _Mode
                  ld de, _Locker
                  ld a, (hl)
                  inc a
                  cp 3
                  jr c, mode_ok
                  ld a, 0
                  .mode_ok
                  ld (hl), a

                  ex de, hl
                  ld a, (hl)
                  cp 0
                  jr nz, mode_ret

                  ex de, hl
                  ld a, (hl)
                  cp 2
                  jr c, mode_ret
                  ld (hl),0

                  .mode_ret
                  call _MenuInfo
              #endasm
              }

          if (p == 50) {  // Ha pulsado el 2
              wyz_play_sound (FX_TECLA);
              if (++TileSet > Locker || TileSet >= MAXTSET)
                 TileSet = 0;           
              InitTileset ();
              MenuInfo(); 
              }

           sp_UpdateNow();

           if (++m > 7) { // Si has desplazado ya 8 pxeles
              m = 0;
              sp_PrintAtInv (23, 31, 0, Roll[Score]); // Pon un nuevo caracter
              if (Roll[++Score] == 0)
                 Score = 0;
              }

           MyScr();
           
           JoyFunc = sp_JoyKeyboard;
           CheckFireIsPressed ();

           JoyFunc = sp_JoyKempston;
           CheckFireIsPressed ();

           JoyFunc = sp_JoySinclair1;
           CheckFireIsPressed ();
           
           JoyFunc = sp_JoySinclair2;
           CheckFireIsPressed ();
           
           }
}

uchar Pause()
{
     wyz_play_sound (FX_PAUSA);
     PrintBanner (5, ORIGENGRIDX, PAPER_YELLOW, "    PAUSED    ");
     PrintBanner (8, ORIGENGRIDX, PAPER_CYAN , "PRESS  ANY KEY");
     PrintBanner (10, ORIGENGRIDX, PAPER_CYAN, "TO RESUME OR G");
     PrintBanner (12, ORIGENGRIDX, PAPER_CYAN, "TO ABORT  GAME");
     sp_UpdateNow();
     sp_WaitForNoKey();
     do {
           Color = sp_GetKey();
           if (Color == 103) // Ha pulsado la G
              return 1;
           Ticks = 0;
           } while (!Color);
     DrawGrid();
     sp_UpdateNow();
     return 0;
}

void ShowHistory(char* Text)
{
     m = 14;             // Al poner el texto empezamos en la posicin 2, 14 y a partir de 14 empezamos en 0  
     n = 1;              // La coordenada vertical la multiplica por dos al ir a imprimir
     asm_int = 0;
     
     while (Text[Counter] != '%') {
           while (Text[Counter] == '$') {
                 Counter++;
                 if (++n > 6)
                    m = 0;
                  else
                    m = 14;
                 }
           for (o = 1; o < 8 && Text[Counter] != ' '; o++) {
               if (!asm_int)
                  asm_int = sp_GetKey() + !(((JoyFunc) (&keys)) & sp_FIRE);
               if (asm_int)
                  o = 7;
               sp_PrintAtInv (n * 2, m, o | BRIGHT, Text[Counter]);
               if (!asm_int) {
               #asm           // Un retardo de menos de 1/50 segs.
                   ld bc, 500
                   .buclenop
                   nop
                   dec bc
                   ld a,b
                   or c
                   jr nz,buclenop
               #endasm
                  sp_UpdateNow();
                  }
               }
           m++;
           Counter++;
           }
     Counter++;
     sp_UpdateNow();
     sp_WaitForNoKey();
}
     
uchar Interlude(uchar Level)
{      
     if ((Level % INTERLEVEL))
        return 0;

     sp_MoveSprAbs (Selector, Tablero, 0, 22,32,0,0);

    if (Level) {

       #asm
/*        for (n = 0; n < DIMGRID * DIMGRID; n++) {
            TileGrid [n] *= -1;
            if (TileGrid [n] < -7)
               TileGrid[n] = -7;
            }
80 bytes menos */
         ld b, 64
         ld hl, _TileGrid
         .bucle_expl
         ld a, (hl)
         cp 7
         jr c, explmenor
         ld a,7
         .explmenor
         neg
         ld (hl), a
         inc hl
         djnz bucle_expl
       #endasm
                        
        if (Locker <= ++TileSet)
           Locker = TileSet;

       InitTileset();
       Explosion(0);
       Fade();
       sp_UpdateNow();
       }
       
     uncrunch ((AddrInterludes [Level / INTERLEVEL]));

     if (Level / INTERLEVEL < 5)
        wyz_play_music (Level / INTERLEVEL);
      else
        wyz_play_music (0);

     ShowHistory (Texts);  
     
     while (!(sp_GetKey() + !(((JoyFunc) (&keys)) & sp_FIRE)));
     Fade();
     wyz_stop_sound();

     if ((Level / INTERLEVEL) >= 5) { // Te acabaste el juego???
        TileSet = 0;
        Locker = ((MAXTSET - 1) << 4) + SPECIALTILE;
        if (is128k) {
           UnpackScreenInPage (3, THEEND);
           wyz_play_music (MUSIC_THEEND);
           sp_WaitForKey();
           Fade();
           wyz_stop_sound();
           }
        return 2; // Game Over "Bueno"
        }
                
     uncrunch (MARCO);
     return 0;
}

void GameOver()
{
     wyz_stop_sound();
     Fade();
     Counter = 0;

     uncrunch (GAMEOVER);

     wyz_play_music (MUSIC_GAMEOVER);
     
     ShowHistory (GOverTxt);

     sp_Pause (0);
     Fade();
     wyz_stop_sound();
}
     
     


void GameCicle ()
{
    uchar i, Salir;
    uchar Ar, Ab, Iz, De, PixelMov;

    Salir = 0;

    while (!Salir) {
         sp_UpdateNow();

          i = (JoyFunc) (&keys);
          
          Ar = i & sp_UP;
          Ab = i & sp_DOWN;
          Iz = i & sp_LEFT;
          De = i & sp_RIGHT;
          
          if (TimeR) {
             sp_Swap (&Ar, &Ab, 1);
             sp_Swap (&Iz, &De, 1);
             }

          if (Iz == 0 && TX > 0)
             if (!Swaping)
                ObjX = TX - 1;
              else {
                SwapTile (TX, TY, -1);
                Swaping = 0;
                }

          if (De == 0 && TX < DIMGRID - 1) 
             if (!Swaping)
                ObjX = TX + 1;
              else {
                SwapTile (TX, TY, 1);
                Swaping = 0;
                }

          if (Ar == 0 && TY > 0)
             if (!Swaping)
                ObjY = TY - 1;
              else {
                SwapTile (TX, TY, -DIMGRID);
                Swaping = 0;
                }

          if (Ab == 0 && TY < DIMGRID - 1) 
             if (!Swaping)
                ObjY = TY + 1;
              else {
                SwapTile (TX, TY, DIMGRID);
                Swaping = 0;
                }

          if ((i & sp_FIRE) == 0 && FireReleased && ObjX == TX && ObjY == TY) {
             FireReleased = 0;
             if (TileGrid [TY * DIMGRID + TX] > 7)
                SpecialTile(TY, TX);
              else {
                Swaping = !Swaping;
                if (Swaping) {
                   wyz_play_sound (FX_GET);
                   DibujaTile (TX, TY, TileGrid [TY * DIMGRID + TX], TileGrid [TY * DIMGRID + TX] << 3, 0, 0);
                   }
                  else
                   DibujaTile (TX, TY, TileGrid [TY * DIMGRID + TX], PAPER_BLACK, 0, 0);
                }
             }

          if ((i & sp_FIRE))
             FireReleased = 1;
          
          PixelMov = CMOV;

          if (Cambiante >= 0) {
             PixelMov *= 2;
             for (n = 0; n < DIMGRID * DIMGRID; n++)
                 if (TileGrid [n] == 18)
                    DibujaTile (n % DIMGRID, n / DIMGRID, 18, PAPER_BLACK, 0, 0);
             }

          if (ObjX < TX)
             XAct -= PixelMov;

          if (ObjX > TX)
             XAct += PixelMov;

          if (ObjY < TY)
             YAct -= PixelMov;

          if (ObjY > TY)
             YAct += PixelMov;

          if (!((XAct - ORIGENGRIDX * 8) % 16))
             TX = (XAct - ORIGENGRIDX) / 16;

          if (!((YAct - ORIGENGRIDY * 8) % 16))
             TY = (YAct - ORIGENGRIDY) / 16;

          sp_MoveSprAbs (Selector, Tablero, 0, YAct / 8, XAct / 8, XAct % 8, YAct % 8);

          if (TimeIndic > 0 && TimeIndic <= Segs)
             ClearIndic();
          if (TimeR <= Segs)
             TimeR = 0;

          ShowTime();

          if (Time - Segs == 10 && !Ticks)
             wyz_play_sound (FX_TIMEATTACK);


          if (sp_GetKey() == 104) // Ha pulsado la H
             Salir = Pause();

          if (!Salir && CheckEnd > 6) {
             Time += TIMEBONUS - Segs;
             if (Time > TIMELIMIT)
                Time = TIMELIMIT;
             if (!Mode)
                Salir = Interlude(Objetivo - 2);
             if (!Salir)
                StartLevel();
             }

          if (!Salir && Time <= Segs) {
             Color = PAPER_RED | INK_YELLOW;
             PrintIndic (0, " GAME", " OVER");
             ShowTime();
             CheckPreLines (1);
             sp_UpdateNow();
             sp_Wait (50);
             sp_Pause (200);
             Salir = 1; // Game Over "Malo"
             if (!Mode)
                GameOver();
             }
    }
    wyz_stop_sound();
}

void InitGame ()
{
    Objetivo = 0;
    Episode = 0;
    Score = 0;
    Counter = 0;
    if (!Mode) {
       TileSet = 0;
       InitTileset();
       Interlude (0);
       }
     else
       uncrunch (MARCO);
    Swaping = 0;
    FireReleased = 1;
    Time = TIMELIMIT;
    TimeIndic = 0;
    StartLevel();
}

void main ()
{
    Init();

    while (1) {
          Menu();
          InitGame();
          GameCicle();
          }
}
